import os
from configparser import ConfigParser

from typing import Any


class EnvConfigParser(ConfigParser):
    ENV_PREFIX = 'RINGR'

    def get(self, section: str, option: str, *args, **kwargs) -> Any:
        env_var = f'{self.ENV_PREFIX}_{section.upper()}_{option.upper()}'
        return os.environ.get(env_var) or super().get(section, option, *args, **kwargs)



import unittest
from unittest.mock import patch

from ringr.config_parser import EnvConfigParser


class ConfigParserTestCase(unittest.TestCase):

    @patch.dict('os.environ', {}, clear=True)
    def test_from_file(self):
        parser = EnvConfigParser()
        parser.read_dict({
            'section1': {
                'value1': '123'
            }
        })

        self.assertEqual('123', parser.get('section1', 'value1'))
        self.assertEqual(123, parser.getint('section1', 'value1'))

    @patch.dict('os.environ', {
        'RINGR_SECTION1_VALUE1': '999'
    }, clear=True)
    def test_env_var_takes_preference(self):
        parser = EnvConfigParser()
        parser.read_dict({
            'section1': {
                'value1': '123'
            }
        })

        self.assertEqual('999', parser.get('section1', 'value1'))
        self.assertEqual(999, parser.getint('section1', 'value1'))

    @patch.dict('os.environ', {}, clear=True)
    def test_fallback(self):
        parser = EnvConfigParser()
        parser.read_dict({
            'section1': {
                'value1': '123'
            }
        })

        self.assertEqual('333', parser.get('section1', 'unknown', fallback='333'))



import os
from pathlib import Path
import logging

from dataclasses import dataclass
from typing import Optional

from .config_parser import EnvConfigParser
from .notifiers import parse_notifier_config, NotifierConfig


log = logging.getLogger(__name__)


@dataclass(frozen=True)
class DetectorConfig:
    device: int
    cooldown_secs: float
    threshold: float
    peak_duration: float
    frequency: int
    block_duration: int
    acceptance_ratio: float = 100
    num_freq_bins: int = 256
    gain: int = 0
    latency: Optional[float] = None
    log_analysis: bool = False

    @classmethod
    def configure(cls, conf: EnvConfigParser):
        return cls(
            device=conf.getint('detector', 'device'),
            cooldown_secs=conf.getfloat('detector', 'cooldown'),
            threshold=conf.getfloat('detector', 'threshold'),
            peak_duration=conf.getfloat('detector', 'peak_duration'),
            frequency=conf.getint('detector', 'frequency'),
            block_duration=conf.getint('detector', 'block_duration'),
            acceptance_ratio=conf.getfloat('detector', 'acceptance_ratio', fallback=cls.acceptance_ratio),
            num_freq_bins=conf.getint('detector', 'frequency_bins', fallback=cls.num_freq_bins),
            gain=conf.getint('detector', 'gain', fallback=cls.gain),
            latency=conf.getfloat('detector', 'latency', fallback=cls.latency),
            log_analysis=conf.getboolean('detector', 'log_analysis', fallback=cls.log_analysis),
        )


@dataclass(frozen=True)
class Config:
    detector: DetectorConfig
    notifier: NotifierConfig


def load_config(file: Path) -> Config:
    parser = EnvConfigParser()
    if os.path.isfile(file):
        parser.read(file)

    notifier_config = parse_notifier_config(parser)
    detector_config = DetectorConfig.configure(parser)
    config = Config(detector_config, notifier_config)

    log.debug('Config used: %s', config)
    return config



import math
import time
import logging

from typing import Any

import sounddevice as sd
import numpy as np

from .config import DetectorConfig
from .notifiers import Notifier


log = logging.getLogger(__name__)


class SlidingWindow:
    def __init__(self, size):
        self.size = size
        self.window = []

    def add(self, value):
        if len(self.window) == self.size:
            self.window.pop(0)
        self.window.append(value)

    def __len__(self):
        return len(self.window)

    def __repr__(self):
        return str(self.window)


class AudioDetector:
    def __init__(self, config: DetectorConfig, notifier: Notifier) -> None:
        self.config = config
        self.notifier = notifier

        self.device = self.config.device
        self.threshold = self.config.threshold / 100.0
        self.peak_duration = self.config.peak_duration
        self.acceptance_ratio = self.config.acceptance_ratio
        self.frequency = self.config.frequency
        self.num_freq_bins = self.config.num_freq_bins
        self.block_duration = self.config.block_duration
        self.gain = self.config.gain
        self.latency = 'high' if self.config.latency is None else self.config.latency

        self.samplerate = self.get_samplerate(self.device)
        self.blocksize = int(self.samplerate * self.block_duration / 1000)
        self.peak_blocks = int(self.samplerate * self.peak_duration / self.blocksize)
        self.acceptable_peak_blocks = self.peak_blocks * self.acceptance_ratio / 100.0
        self.sliding_window = SlidingWindow(self.peak_blocks)
        self.last_detection_time = 0

        # Notes:
        #   samplerate / 2: maximum frequency that can be correctly captured
        #   num_freq_bins: number of bins for the fft
        #   fftsize: size of the fft bin
        #   freq_bin_idx: bin index for the desired frequency to analyze
        max_freq = self.samplerate / 2
        delta_f = max_freq / (self.num_freq_bins - 1)
        self.fftsize = math.ceil(self.samplerate / delta_f)
        self.freq_bin_idx = math.ceil(self.frequency / delta_f)

        self.last_state = None
        self.update_state(False)

    @staticmethod
    def get_samplerate(device: int) -> int:
        """ Get default samplerate of the input sound device """
        return sd.query_devices(device, 'input')['default_samplerate']

    def start(self) -> None:
        with sd.InputStream(
            device=self.device,
            channels=1,
            samplerate=self.samplerate,
            blocksize=self.blocksize,
            latency=self.latency,
            callback=self.callback
        ):
            while True:
                time.sleep(1)

    def callback(self, indata: np.ndarray, frames: int, stime: Any, status: sd.CallbackFlags) -> None:
        """This is called (from a separate thread) for each audio block."""
        if status:
            log.error('Error status: %s', status)
        if any(indata):
            self.analyze(indata)
        else:
            log.debug('No input')

    def analyze(self, data: np.ndarray) -> None:
        magnitude = self.get_fft_magnitude(data)
        detected = self.process_value(magnitude)
        # Cooldown reset check
        if self.last_state:
            if (time.time() - self.last_detection_time) < self.config.cooldown_secs:
                # Do nothing during cooldown time
                return
            else:
                self.update_state(False)
        # Detection
        if detected:
            log.info('Sound event detected')
            self.last_detection_time = time.time()
            self.update_state(True)

    def get_fft_magnitude(self, data: np.ndarray) -> float:
        magnitudes = np.abs(np.fft.rfft(data[:, 0], n=self.fftsize))
        magnitude = magnitudes[self.freq_bin_idx]
        magnitude *= self.gain / self.fftsize
        magnitude = np.clip(magnitude, 0, 1)  # normalized between 0 and 1, limit values
        return magnitude

    def process_value(self, value: float) -> bool:
        matches = value > self.threshold
        self.sliding_window.add(matches)
        if len(self.sliding_window) < self.peak_blocks:
            # We need more samples to take a decision
            return False
        num_matches = self.sliding_window.window.count(True)
        if self.config.log_analysis:
            log.debug('Value %s - Threshold %s - num_matches %s - acceptable %s', value, self.threshold,
                      num_matches, self.acceptable_peak_blocks)
        return num_matches >= self.acceptable_peak_blocks

    def update_state(self, new_state: bool):
        self.last_state = new_state
        self.notifier.notify(new_state)



import unittest
from unittest.mock import patch

from ringr.config_parser import EnvConfigParser
from ringr.config import DetectorConfig


class ConfigTestCase(unittest.TestCase):

    @patch.dict('os.environ', {}, clear=True)
    def test_detector_config(self):
        parser = EnvConfigParser()
        parser.read_dict({
            'detector': {
                'device': '1',
                'cooldown': '10',
                'threshold': '60',
                'peak_duration': '0.8',
                'acceptance_ratio': '0.85',
                'frequency': '1000',
                'frequency_bins': '256',
                'gain': '200',
                'block_duration': '50',
                'latency': '0.1',
                'log_analysis': 'True',
            }
        })

        detector_config = DetectorConfig.configure(parser)
        expected = DetectorConfig(
            device=1,
            cooldown_secs=10,
            threshold=60,
            peak_duration=0.8,
            acceptance_ratio=0.85,
            frequency=1000,
            num_freq_bins=256,
            gain=200,
            block_duration=50,
            latency=0.1,
            log_analysis=True
        )
        self.assertEqual(expected, detector_config)

    @patch.dict('os.environ', {
        'RINGR_DETECTOR_DEVICE': '1',
        'RINGR_DETECTOR_COOLDOWN': '10',
        'RINGR_DETECTOR_THRESHOLD': '60',
        'RINGR_DETECTOR_PEAK_DURATION': '0.8',
        'RINGR_DETECTOR_ACCEPTANCE_RATIO': '0.85',
        'RINGR_DETECTOR_FREQUENCY': '1000',
        'RINGR_DETECTOR_FREQUENCY_BINS': '256',
        'RINGR_DETECTOR_GAIN': '200',
        'RINGR_DETECTOR_BLOCK_DURATION': '50',
        'RINGR_DETECTOR_LATENCY': '0.1',
        'RINGR_DETECTOR_LOG_ANALYSIS': 'True',
    }, clear=True)
    def test_detector_config_from_env(self):
        parser = EnvConfigParser()
        parser['detector'] = {}
        parser['detector']['device'] = '2'

        detector_config = DetectorConfig.configure(parser)
        expected = DetectorConfig(
            device=1,
            cooldown_secs=10,
            threshold=60,
            peak_duration=0.8,
            acceptance_ratio=0.85,
            frequency=1000,
            num_freq_bins=256,
            gain=200,
            block_duration=50,
            latency=0.1,
            log_analysis=True
        )
        self.assertEqual(expected, detector_config)

    @patch.dict('os.environ', {}, clear=True)
    def test_optional_params(self):
        parser = EnvConfigParser()
        parser.read_dict({
            'detector': {
                'device': '1',
                'cooldown': '10',
                'threshold': '60',
                'peak_duration': '0.8',
                'acceptance_ratio': '0.85',
                'frequency': '1000',
                'frequency_bins': '256',
                'gain': '200',
                'block_duration': '50',
            }
        })

        detector_config = DetectorConfig.configure(parser)
        expected = DetectorConfig(
            device=1,
            cooldown_secs=10,
            threshold=60,
            peak_duration=0.8,
            acceptance_ratio=0.85,
            frequency=1000,
            num_freq_bins=256,
            gain=200,
            block_duration=50,
            latency=None,
            log_analysis=False
        )
        self.assertEqual(expected, detector_config)



import unittest
from unittest.mock import patch


from ringr.notifiers.telegram_notifier import TelegramNotifier, TelegramNotifierConfig


class TelegramNotifierTestCase(unittest.TestCase):
    config = TelegramNotifierConfig(
        type='telegram',
        api_token=':my_api_token:',
        chat_id='my_chat_id',
        message='the message'
    )

    def setUp(self):
        urllib_patcher = patch('ringr.notifiers.telegram_notifier.urllib.request')
        self.mock_request = urllib_patcher.start()
        self.addCleanup(urllib_patcher.stop)

        self.notifier = TelegramNotifier(self.config)

    def test_notify_detected(self):
        self.notifier.notify(True)

        expected_url = 'https://api.telegram.org/bot:my_api_token:/sendMessage'
        expected_payload = '{"chat_id": "my_chat_id", "text": "the message"}'.encode()
        expected_headers = {'Content-Type': 'application/json'}
        self.mock_request.Request.assert_called_once_with(expected_url, expected_payload,
                                                          headers=expected_headers)

        self.mock_request.urlopen.assert_called_once()

    def test_notify_undetected(self):
        self.notifier.notify(False)

        self.mock_request.urlopen.assert_not_called()



import unittest
from unittest.mock import Mock, MagicMock, patch

import numpy as np

from ringr.audio import AudioDetector
from ringr.config import DetectorConfig


class AudioDetectorTestCase(unittest.TestCase):
    config = DetectorConfig(
        device=1,
        cooldown_secs=10,
        threshold=65,
        peak_duration=1.5,
        frequency=1000,
        block_duration=50,
        acceptance_ratio=95,
        num_freq_bins=256,
        gain=200,
        latency=0.1,
        log_analysis=False
    )

    def setUp(self):
        self.notifier = Mock()

        AudioDetector.get_samplerate = Mock(return_value=44100)

        self.sliding_window = MagicMock()

        self.detector = AudioDetector(self.config, self.notifier)
        self.detector.sliding_window = self.sliding_window

        self.data = np.full([2205, 2], 1000)  # blocksize

    def test_initial_parameters(self):
        self.assertEqual(1, self.detector.device)
        self.assertEqual(44100, self.detector.samplerate)
        self.assertEqual(0.1, self.detector.latency)
        self.assertEqual(50, self.detector.block_duration)
        self.assertEqual(2205, self.detector.blocksize)  # 50 ms at 44100 samples per second
        self.assertEqual(0.65, self.detector.threshold)
        self.assertEqual(1.5, self.detector.peak_duration)
        self.assertEqual(30, self.detector.peak_blocks)  # TODO: review
        self.assertEqual(95, self.detector.acceptance_ratio)
        self.assertEqual(28.5, self.detector.acceptable_peak_blocks)
        self.assertEqual(256, self.detector.num_freq_bins)
        self.assertEqual(200, self.detector.gain)

        # FFT
        self.assertEqual(510, self.detector.fftsize)
        self.assertEqual(12, self.detector.freq_bin_idx)

    def test_fft_filter(self):
        # This does not test the actual FFT calculation, this is delegated to numpy, but it tests there is
        # no undesired side effect or exception thrown
        self.assertAlmostEqual(9.634e-13, self.detector.get_fft_magnitude(self.data), delta=0.001)

    def test_process_value_not_enough_samples(self):
        # Fast exit if there is not enough samples in the sliding window
        self.sliding_window.__len__.return_value = 1

        self.assertFalse(self.detector.process_value(1))
        self.sliding_window.add.assert_called_once_with(True)
        self.sliding_window.window.count.assert_not_called()

    def test_process_value_match(self):
        self.sliding_window.__len__.return_value = 31   # higher than peak blocks. TODO: review
        self.sliding_window.window.count.return_value = 29

        self.assertTrue(self.detector.process_value(1))

    def test_process_value_under_acceptance_ratio(self):
        self.sliding_window.__len__.return_value = 31   # higher than peak blocks. TODO: review
        self.sliding_window.window.count.return_value = 28

        self.assertFalse(self.detector.process_value(1))

    def test_process_value_added_sample_under_threshold(self):
        self.detector.process_value(0.64)
        self.sliding_window.add.assert_called_once_with(False)

    def test_process_value_added_sample_equal_to_threshold(self):
        self.detector.process_value(0.65)
        self.sliding_window.add.assert_called_once_with(False)

    def test_process_value_added_sample_above_threshold(self):
        self.detector.process_value(0.66)
        self.sliding_window.add.assert_called_once_with(True)

    @patch('ringr.audio.time')
    def test_analyze_not_detected(self, mock_time):
        self.detector.last_detection_time = 0
        self.detector.last_state = False
        mock_time.time.return_value = 15
        self.notifier.notify.reset_mock()

        self.detector.get_fft_magnitude = Mock()
        self.detector.process_value = Mock(return_value=False)

        self.detector.analyze(self.data)

        self.assertFalse(self.detector.last_state)
        self.notifier.notify.assert_not_called()

    @patch('ringr.audio.time')
    def test_analyze_detected(self, mock_time):
        self.detector.last_detection_time = 0
        self.detector.last_state = False
        mock_time.time.return_value = 15
        self.notifier.notify.reset_mock()

        self.detector.get_fft_magnitude = Mock()
        self.detector.process_value = Mock(return_value=True)

        self.detector.analyze(self.data)

        self.assertTrue(self.detector.last_state)
        self.assertEqual(15, self.detector.last_detection_time)
        self.notifier.notify.assert_called_once_with(True)

    @patch('ringr.audio.time')
    def test_analyze_cooldown_time(self, mock_time):
        self.detector.last_detection_time = 0
        self.detector.last_state = True
        mock_time.time.return_value = 9
        self.notifier.notify.reset_mock()

        self.detector.get_fft_magnitude = Mock()
        self.detector.process_value = Mock()

        self.detector.analyze(self.data)

        self.assertTrue(self.detector.last_state)
        self.notifier.notify.assert_not_called()
        self.detector.get_fft_magnitude.assert_not_called()
        self.detector.process_value.assert_not_called()

    @patch('ringr.audio.time')
    def test_analyze_reset_cooldown_time(self, mock_time):
        self.detector.last_detection_time = 0
        self.detector.last_state = True
        mock_time.time.return_value = 16
        self.notifier.notify.reset_mock()

        self.detector.get_fft_magnitude = Mock()
        self.detector.process_value = Mock(return_value=False)

        self.detector.analyze(self.data)

        self.assertFalse(self.detector.last_state)
        self.notifier.notify.assert_called_once_with(False)



import unittest
from unittest.mock import Mock, patch

from ringr.notifiers.ha_notifier import HANotifier, HANotifierConfig


class HANotifierTestCase(unittest.TestCase):
    config = HANotifierConfig(
        type='ha',
        device_id='ringr_01',
        device_name='ringr 01',
        mqtt_host='localhost',
        mqtt_port=1883,
        mqtt_user='ringr_user',
        mqtt_pass='ringr_pass',
        mqtt_client_id='ringr_mqtt_01',
        mqtt_qos=1
    )

    def setUp(self):
        paho_patcher = patch('ringr.notifiers.ha_notifier.paho')
        self.mock_paho = paho_patcher.start()
        self.addCleanup(paho_patcher.stop)

        threading_patcher = patch('ringr.notifiers.ha_notifier.threading')
        self.mock_threading = threading_patcher.start()
        self.addCleanup(threading_patcher.stop)

        self.connected_event = Mock()
        self.mock_threading.Event = Mock(return_value=self.connected_event)

        self.mqtt = Mock()
        self.mock_paho.Client = Mock(return_value=self.mqtt)
        self.mock_paho.CONNACK_ACCEPTED = 0

        self.notifier = HANotifier(self.config)

    def test_connect(self):
        self.mock_paho.Client.assert_called_once_with(client_id='ringr_mqtt_01')
        self.mqtt.username_pw_set.asset_called_once_with(username='ringr_user', password='ringr_pass')
        self.mqtt.connect.assert_called_once_with(host='localhost', port=1883)
        self.mqtt.loop_start.assert_called_once()
        self.connected_event.wait.assert_called_once()

    def test_connect_no_auth(self):
        config = HANotifierConfig(
            type='ha',
            device_id='ringr_01',
            device_name='ringr 01',
            mqtt_host='localhost',
            mqtt_port=1883,
            mqtt_client_id='ringr_mqtt_01',
            mqtt_qos=1
        )

        self.mqtt = Mock()
        self.mock_paho.Client = Mock(return_value=self.mqtt)

        self.notifier = HANotifier(config)

        self.mqtt.username_pw_set.assert_not_called()

    def test_on_connect_accepted(self):
        self.notifier._on_mqtt_connect(None, None, None, 0)

        self.connected_event.set.assert_called_once()

    def test_on_connect_error(self):
        self.notifier._on_mqtt_connect(None, None, None, 1)

        self.connected_event.set.assert_not_called()

    def test_publish_discovery_config_message_on_connect(self):
        expected_topic = 'homeassistant/binary_sensor/ringr_01/config'
        expected_payload = ('{"name": "ringr_01", '
                            '"unique_id": "ringr_01", '
                            '"device": {"name": "ringr 01", "identifiers": ["ringr_01"]}, '
                            '"manufacturer": "Alberto Alcolea", '
                            '"model": "ringr", '
                            '"sw_version": "0.1.0", '
                            '"device_class": "sound", '
                            '"state_topic": "homeassistant/binary_sensor/ringr_01/state", '
                            '"value_template": "{{ value_json.state }}"}')

        self.mqtt.publish.assert_called_once_with(expected_topic, expected_payload, retain=True, qos=1)

    def test_notify_detected(self):
        self.notifier.notify(True)

        expected_topic = 'homeassistant/binary_sensor/ringr_01/state'
        expected_payload = '{"state": "ON"}'

        self.mqtt.publish.assert_called_with(expected_topic, expected_payload, retain=True, qos=1)

    def test_notify_undetected(self):
        self.notifier.notify(False)

        expected_topic = 'homeassistant/binary_sensor/ringr_01/state'
        expected_payload = '{"state": "OFF"}'

        self.mqtt.publish.assert_called_with(expected_topic, expected_payload, retain=True, qos=1)



import sys
import argparse
from pathlib import Path
import logging

from .config import load_config
from .notifiers import create_notifier
from .audio import AudioDetector


log = logging.getLogger(__name__)


DEFAULT_CONFIG_FILE = '/etc/ringr/ringr.conf'


def configure_logger(verbose: bool):
    log.setLevel(logging.DEBUG if verbose else logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    log.addHandler(handler)


def parse_args():
    parser = argparse.ArgumentParser(description='ringr. Sound event detection system')
    parser.add_argument('-c', '--conf', help='Configuration file', metavar='file',
                        default=DEFAULT_CONFIG_FILE)
    parser.add_argument('-v', '--verbose', help='Show debug log messages in the log', action='store_true')
    return parser.parse_args()


def main():
    args = parse_args()
    configure_logger(args.verbose)

    try:
        config = load_config(Path(args.conf))
        notifier = create_notifier(config.notifier)
        detector = AudioDetector(config.detector, notifier)

        log.info('Starting detector')

        detector.start()

    except (KeyboardInterrupt, SystemExit):
        # Do nothing
        pass

    except Exception:
        log.error('Something went wrong', exc_info=True)
        sys.exit(1)


if __name__ == '__main__':
    main()



import re

from setuptools import setup


PACKAGE_NAME = 'ringr'


with open('requirements.txt', 'r') as f:
    requires = f.read().split()


def read_file(path):
    with open(path) as f:
        return f.read()


def get_version():
    content = read_file(f'{PACKAGE_NAME}/__init__.py')
    regex = r'^__version__ = [\'"]([^\'"]+)[\'"]'
    match = re.search(regex, content, re.MULTILINE)
    if match:
        return match.group(1)
    raise RuntimeError('Unable to find __version__')


setup(
    name=PACKAGE_NAME,
    version=get_version(),
    description='Sound event detection system',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    author='Alberto Alcolea',
    author_email='albertoalcolea@gmail.com',
    url='https://github.com/albertoalcolea/ringr',
    project_urls={
        'Source': 'https://github.com/albertoalcolea/ringr'
    },
    license='MIT',
    keywords=[
        'ringr',
        'ring',
        'sound',
        'sound detection',
        'audio',
        'audio detection',
        'frequency analyzer',
        'peak detection',
        'home assistant',
    ],
    packages=[PACKAGE_NAME],
    install_requires=requires,
    entry_points={
        'console_scripts': [
            'ringr = ringr.__main__:main',
        ]
    },
    extras_require={
        'dev': [
            'flake8>=4.0.1',
            'twine>=4.0.0',
        ]
    },
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
    ],
)
